# Copyright (c) 2022-2023 The pymovements Project Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module holds utils for developers and is not part of the user API.
"""
from __future__ import annotations

from typing import Any
from typing import TypeVar

ClassT = TypeVar('ClassT')


def auto_str(cls: type[ClassT]) -> type[ClassT]:
    """
    Automatically generate __str__() to include all arguments. Can be used as a decorator.
    """
    def shorten(value: Any) -> str:
        if isinstance(value, float):
            value = f'{value:.2f}'
        return value

    def __str__(self: Any) -> str:
        attributes = ', '.join(f'{key}={shorten(value)}' for key, value in vars(self).items())
        return f'{type(self).__name__}({attributes})'

    # for type ignore see: https://github.com/python/mypy/issues/3951#issuecomment-329183108
    cls.__str__ = __str__  # type: ignore
    return cls
